#!/usr/bin/env python

# avahihelper - helper classes for publishing/browsing avahi services!
#
# Copyright (C) 2008, Aldrin Martoq <aldrin@martoq.cl>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import dbus, gobject, avahi

def trace(fn):
    "Returns a traced version of the input function."
    from itertools import chain
    from time import localtime, strftime
    def decorator(*v, **k):
        name = fn.__name__
        print "%s %s.%s(%s)" % (
            strftime("%Y%m%d %H%M%S", localtime()),
            repr(v[0]), name, ", ".join(map(repr, chain(v[1:], k.values()))))
        return fn(*v, **k)
    return decorator

class AvahiService():
    def __init__(self, name = None, stype = None, port = None, txt = [],
                 domain = "", host = "", iface = avahi.IF_UNSPEC, 
                 addr = None, proto = avahi.PROTO_UNSPEC, aproto = None,
                 flags = 0, rcount = 10):
        self.name = name
        self.stype = stype
        self.port = port
        self.txt = txt
        self.domain = domain
        self.host = host
        self.iface = iface
        self.addr = addr
        self.proto = proto
        self.aproto = aproto
        self.flags = flags
        self.rcount = rcount
    
    def __cmp__(self, other):
        for name in ['stype', 'name', 'proto', 'iface', 'domain']:
            c = cmp(getattr(self, name), getattr(other, name))
            if c != 0: return c
        return 0
    
    def __str__(self):
        r = "<%s iface=%s proto=%s name='%s' type='%s' domain='%s' " \
            "host='%s' aproto=%s addr=%s port=%s txt=%s flags=%s>" % \
            (self.__class__.__name__,
             self.iface,
             self.proto,
             self.name,
             self.stype,
             self.domain,
             self.host,
             self.aproto,
             self.addr,
             self.port,
             self.txt,
             self.flags)
        return r
    
    def __repr__(self):
        r = "<%s 0x%x>" % \
            (self.__class__.__name__,
             id(self))
        return r

### CAMBIAR PARA QUE FUNCIONE
###class AvahiBrowser(gobject.Gobject):
class AvahiBrowser(gobject.GObject, list):
    __gsignals__ = {
        "started" : (gobject.SIGNAL_RUN_FIRST, None, ()),
        "stopped" : (gobject.SIGNAL_RUN_FIRST, None, (object,)),
        "found" : (gobject.SIGNAL_RUN_FIRST, None, (object,)),
        "gone" : (gobject.SIGNAL_RUN_FIRST, None, (object,))
    }
    
    @trace
    def __init__(self, bus = None, stype = '', ignore_local = True, domain='local'):
        gobject.GObject.__init__(self)
        self.bus = bus
        self.stype = stype
        self.ignore_local = ignore_local
        self.server = None
        self.browser = None
        self.domain = domain
        self.items = []

    def len(self):
        return len(self.items)
    
    @trace
    def browse(self):
        if not self.server is None:
            print "Warning! already started!"
            return
        obj = self.bus.get_object(avahi.DBUS_NAME, avahi.DBUS_PATH_SERVER)
        self.server = dbus.Interface(obj, avahi.DBUS_INTERFACE_SERVER)
        brw = self.server.ServiceBrowserNew(avahi.IF_UNSPEC,
                                            avahi.PROTO_UNSPEC,
                                            self.stype,
                                            self.domain,
                                            dbus.UInt32(0))
        obj = self.bus.get_object(avahi.DBUS_NAME, brw)
        self.browser = dbus.Interface(obj, avahi.DBUS_INTERFACE_SERVICE_BROWSER)
        self.browser.signal_n = self.browser.connect_to_signal('ItemNew', self._browser_itemnew)
        self.browser.signal_r = self.browser.connect_to_signal('ItemRemove', self._browser_itemremove)
        self.emit('started')
        
    @trace
    def unbrowse(self):
        self.browser.signal_n.remove()
        self.browser.signal_r.remove()
        self.browser = None
        self.server = None
        self.emit('stopped')
    
    @trace
    def _browser_itemnew(self, iface, proto, name, stype, domain, flags):
        if self.ignore_local:
            if flags & avahi.LOOKUP_RESULT_LOCAL: return
        
        self.server.ResolveService(iface, proto, name, stype, domain, 
                                   avahi.PROTO_UNSPEC, dbus.UInt32(0),
                                   reply_handler=self._server_resolved,
                                   error_handler=self._server_resolveerror,
                                   byte_arrays=True)
    
    @trace
    def _browser_itemremove(self, iface, proto, name, stype, domain, flags):
        t = AvahiService(iface=iface, proto=proto, name=name,
                         stype=stype, domain=domain)
        if t in self.items:
            s = self.items[self.items.index(t)]
            self.items.remove(s)
            self.emit('gone', s)
        
    @trace
    def _server_resolved(self, *args):
        iface, proto, name, stype, domain, host, \
        aproto, addr, port, text, flags = args
        txt = []
        for t in text: txt.append(str(t))
        s = AvahiService(name, stype, port, txt, domain, host, iface,
                         addr, proto, aproto, flags)
        if not s in self.items:
            self.items.append(s)
            self.emit('found', s)
        
    @trace
    def _server_resolveerror(self, *args):
        print "FIXME: Error on ResolveService: " + str(args[0])
        
    def __str__(self):
        r = ["<%s type='%s' domain='%s' ignore_local=%s" % \
            (self.__class__.__name__,
             self.stype,
             self.domain,
             self.ignore_local)]
        for s in self.items:
            r.append("\t" + str(s))
        r.append("</%s>" % self.__class__.__name__)
        return "\n".join(r)
        

    def __repr__(self):
        r = "<%s 0x%x>" % \
            (self.__class__.__name__,
             id(self))
        return r

class AvahiPublisher():
    @trace
    def __init__(self, bus, service):
        self.bus = bus
        self.service = service
        if self.service.rcount is None:
            self.service.rcount = 12
        self.server = None
        self.group = None

    @trace
    def publish(self):
        if not self.server is None:
            print "Warning! already started!"
            return
        o = self.bus.get_object(avahi.DBUS_NAME, avahi.DBUS_PATH_SERVER)
        self.server = dbus.Interface(o, avahi.DBUS_INTERFACE_SERVER)
        self._server_statechanged(self.server.GetState())
        
    @trace
    def unpublish(self):
        self._del_service()
        pass
    
    @trace
    def _add_service(self):
        if self.group is None:
            obj = self.bus.get_object(avahi.DBUS_NAME, self.server.EntryGroupNew())
            self.group = dbus.Interface(obj, avahi.DBUS_INTERFACE_ENTRY_GROUP)
            self.group.connect_to_signal('StateChanged', self._group_statechanged)
        
        s = self.service
        t = avahi.string_array_to_txt_array(s.txt)
        try:
            self.group.AddService(s.iface,              # iface
                                  s.proto,              # proto
                                  dbus.UInt32(s.flags), # flags
                                  s.name,               # name
                                  s.stype,              # type
                                  s.domain,             # domain
                                  s.host,               # host
                                  dbus.UInt16(s.port),  # port
                                  t                     # txt array
                                  )
            self.group.Commit()
        except dbus.DBusException, e:
            if e._dbus_error_name == 'org.freedesktop.Avahi.CollisionError':
                self._try_alternate_name()
            else:
                raise e
    
    @trace
    def _try_alternate_name(self):
        s = self.service
        s.rcount -= 1
        if s.rcount > 0:
            s.name = self.server.GetAlternativeServiceName(s.name)
            self._del_service()
            self._add_service()
        else:
            print "FIXME: emit signal: Collision: too many changes :("
        pass

    @trace
    def _del_service(self):
        if not self.group is None:
            self.group.Reset()
    
    @trace
    def _group_statechanged(self, state, error):
        if state == avahi.ENTRY_GROUP_ESTABLISHED:
            print "Service published!: %s" % (self.service)
            pass
        elif state == avahi.ENTRY_GROUP_COLLISION:
            self._try_alternate_name()

    @trace
    def _server_statechanged(self, state):
        if state == avahi.SERVER_COLLISION:
            self._del_service()
            print "FIXME: emit signal"
        elif state == avahi.SERVER_RUNNING:
            self._add_service()

    def __repr__(self):
        r = "<%s 0x%x>" % \
            (self.__class__.__name__,
             id(self))
        return r
